module Day15Lib
  ( Vector (..),
    Sensor (..),
    parseNumber,
    parseLine,
    generateBoundary,
    --
    processInput1,
    processInput2,
  )
where

import Data.Bifunctor (Bifunctor (bimap))
import Data.List
import qualified Data.Set as Set
import Debug.Trace
import Text.Parsec

type Parser = Parsec String ()

data Sensor = Sensor
  { position :: Vector,
    beacon :: Vector
  }
  deriving (Show, Eq)

newtype Vector = Vector (Int, Int) deriving (Show, Eq, Ord)

instance Num Vector where
  Vector v1 + Vector v2 = Vector (fst v1 + fst v2, snd v1 + snd v2)
  Vector v1 - Vector v2 = Vector (fst v1 - fst v2, snd v1 - snd v2)
  Vector v1 * Vector v2 = Vector (fst v1 * fst v2, snd v1 * snd v2)
  fromInteger i = Vector (fromInteger i, fromInteger i)
  abs (Vector v) = Vector (bimap abs abs v)
  signum (Vector (x, y)) = Vector (signum x, signum y)

l1d :: Vector -> Vector -> Int
l1d v1 v2 = dx + dy where Vector (dx, dy) = abs (v1 - v2)

sensorRange :: Sensor -> Int
sensorRange sensor = l1d (position sensor) (beacon sensor)

parseSensor :: Parser Sensor
parseSensor = do
  string "Sensor at x="
  sx <- parseNumber
  string ", y="
  sy <- parseNumber
  string ": closest beacon is at x="
  bx <- parseNumber
  string ", y="
  by <- parseNumber
  return $
    Sensor
      { position = Vector (sx, sy),
        beacon = Vector (bx, by)
      }

parseNumber :: Parser Int
parseNumber = do
  sign <- option "" (string "-")
  number <- many1 digit
  return . read $ sign ++ number

parseLine :: String -> Sensor
parseLine line = case parse parseSensor "" line of
  Left _ -> error $ "parse error: " ++ line
  Right s -> s

-- ####################

processInput1 :: Int -> String -> Int
processInput1 ylevel input =
  let sensors = map parseLine . lines $ input
      beaconsInLine =
        Set.fromList
          . map (\Sensor {beacon = Vector (x, _)} -> x)
          . filter (\Sensor {beacon = Vector (_, y)} -> y == ylevel)
          $ sensors
      relevant = filter (isRelevant ylevel) sensors
      empte = foldl Set.union Set.empty . map (Set.fromList . generateEmpty ylevel) $ relevant
   in Set.size $ Set.difference empte beaconsInLine

generateEmpty :: Int -> Sensor -> [Int]
generateEmpty ylevel s@(Sensor {position = Vector (x, y)}) = [x - d .. x + d] where d = sensorRange s - abs (y - ylevel)

isRelevant :: Int -> Sensor -> Bool
isRelevant ylevel s@(Sensor {position = Vector (_, y)}) = abs (y - ylevel) <= sensorRange s

processInput2 :: Int -> String -> Int
processInput2 r input =
  let sensors = map parseLine . lines $ input
      candidates =
        filter (\(Vector (x, y)) -> x >= 0 && x <= r && y >= 0 && y <= r) $
          concatMap generateBoundary sensors
      [Vector (x, y)] = nub $ [c | c <- candidates, not $ any (isDetected c) sensors]
   in x * 4000000 + y

isDetected :: Vector -> Sensor -> Bool
isDetected v s = l1d (position s) v <= sensorRange s

generateBoundary :: Sensor -> [Vector]
generateBoundary s@(Sensor {position = Vector (x, y)}) =
  map
    (position s +)
    ( zipWith (curry Vector) [(-d) .. d] ([0 .. d] ++ [(d - 1), (d - 2) .. 0])
        ++ zipWith (curry Vector) [(-d + 1) .. (d - 1)] ([-1, -2 .. (-d)] ++ [(-d + 1), (-d + 2) .. 0])
    )
  where
    d = sensorRange s + 1
